@@ -67,6 +67,7 @@
 #include <stdarg.h>
 #include <ctype.h>
 #include "sqlite3.h"
+#define ISDIGIT(X) isdigit((unsigned char)(X))
 
 /*
 ** All global variables are gathered into the "g" singleton.
@@ -322,6 +323,7 @@ static void showHelp(void){
 "Options:\n"
 "  --autovacuum          Enable AUTOVACUUM mode\n"
 "  --database FILE       Use database FILE instead of an in-memory database\n"
+"  --disable-lookaside   Turn off lookaside memory\n"
 "  --heap SZ MIN         Memory allocator uses SZ bytes & min allocation MIN\n"
 "  --help                Show this help text\n"    
 "  --lookaside N SZ      Configure lookaside for N slots of SZ bytes each\n"
@@ -382,7 +384,7 @@ static int integerValue(const char *zArg){
       zArg++;
     }
   }else{
-    while( isdigit(zArg[0]) ){
+    while( ISDIGIT(zArg[0]) ){
       v = v*10 + zArg[0] - '0';
       zArg++;
     }
@@ -457,6 +459,7 @@ int main(int argc, char **argv){
   const char *zDbName = 0;      /* Name of an on-disk database file to open */
 
   iBegin = timeOfDay();
+  sqlite3_shutdown();
   zFailCode = getenv("TEST_FAILURE");
   g.zArgv0 = argv[0];
   zPrompt = "<stdin>";
@@ -473,6 +476,10 @@ int main(int argc, char **argv){
         zDbName = argv[i+1];
         i += 1;
       }else
+      if( strcmp(z,"disable-lookaside")==0 ){
+        nLook = 1;
+        szLook = 0;
+      }else
       if( strcmp(z, "f")==0 && i+1<argc ){
         i++;
         goto addNewInFile;
